//**********************************************************************************
// Tango object interface:
//==================================================================================
// constructors:
//  TNode(AppletBaseInterface)
// methods:
//  public 	 void init();
//	issues a request for synchronizing with the session
//  public final boolean isMaster();
//	checks if the current user is the master of the session
//  public final String  getUserName();
//	returns the name of the current user
//  public final String  getMasterName();
//	returns the name of the master
//  public final String  getHostName();
//	returns the name of the local host
//  public final Vector  getParticipants();
//	returns the list of participants names
//  public final Vector  getLocalApps();
// 	returns the list of local applications
//  public final void 	receive(messages.AppEventMessage);
//	this method is to be called from the class extending AppletBase,
//  protected 	 void 	userEntered(String);
//	called whenever a user Enters the session
//  protected 	 void 	userLeft(String);
//	called whenever a user leaves the session
//  protected 	 void 	setMaster(boolean);
//	called whenever master is changed
//  public 	 MessageOutputStream getMos(Object o,boolean isRemovable);
//	gets the stream for messages. If o is TObject the message will be routed
//	to corresponding TObject in all applications in the session, otherwise it
//	will go to receiveData method.
//  protected 	 void 	receiveData(DataInputStream);
//	handles data that is not handled by TObjects
//  protected 	 void 	writeInit(DataOutputStream dos);
//	this should be overriden to write initilization information, which should
//	be passed before sending all TObjects
//  protected 	 void 	readInit(DataInputStream dis);
//	this thouls be overriden to read initialization info written by writeInit
//  public final void 	add(TObject);
//	adds a new TObject with automaticly generated key, corresponding TObject
//	will be created on all applications in the session
//  public final void   add(String key, TObject to);
//	adds a new TObject with a given key. This may be used for creating of
//	predefined objects.
//  public final void 	remove(TObject);
//	remove a TObject, removes a TObject, and destroys all corresponding TObjects
//	in the session
//  public 	 TObject	getTObject(String key);
//  public 	 TObject	getKey(TObject to);
//  public 	 String 	newId();
//**********************************************************************************



package tango;

import java.lang.reflect.*;
import java.io.*;
import java.util.*;
import messages.*;
import consts.*;

public class TNode implements Runnable
{
  AppletBaseInterface abi;
  private Semo s_isMaster;
  private Semo s_masterName;
  private Semo s_userName;
  private Semo s_hostName;
  private Semo s_participants;
  private Semo s_activeUsers;
  private Semo s_localApps;
  private Thread th;
  private Vector msgQueue;

  private final static byte TOBJECT_CRE=0;
  private final static byte TOBJECT_DEST=1;
  private final static byte TOBJECT_DAT=2;
  private final static byte NON_TOBJECT_DAT=3;
  private final static byte INIT_REQ=4;
  private final static byte INIT_RES=5;
  private final static byte LAST_CODE=5;

  private Hashtable tobject;
  private Hashtable trobject;
  private boolean initialized;
  
  private AppEventMessage aem;
  private boolean running;
  private Object  thsync;

  public TNode(AppletBaseInterface abi)
    {
      this.abi=abi;
      initialized=false;
      thsync=new Object();
      s_isMaster=new Semo(this,CAConsts.ISMASTER);
      s_masterName=new Semo(this,CAConsts.MASTERNAME);
      s_userName=new Semo(this,CAConsts.USERNAME);
      s_hostName=new Semo(this,CAConsts.HOSTNAME);
      s_participants=new Semo(this,CAConsts.PARTICIPANTS);
      s_activeUsers=new Semo(this,CAConsts.ACTIVEUSERS);
      s_localApps=new Semo(this,CAConsts.LOCALAPPS);
      tobject=new Hashtable();
      trobject=new Hashtable();
      running=false;
      msgQueue=new Vector();
    }
  public void init()
    {
      v_participants=getParticipants();
      v_isMaster=isMaster();
      
      if(isMaster())
	initialized=true;
      else
	try
	  {
	    MessageOutputStream mos=new MessageOutputStream(this,false,new ByteArrayOutputStream());
	    mos.writeBoolean(false);
	    mos.writeByte(INIT_REQ);
	    mos.send();
	  }
      catch(Exception e)
	{
	  e.printStackTrace();
	}
    }

  public final AppEventMessage messageNext()
    {
      AppEventMessage aem;
      synchronized(msgQueue)
	{
	  try
	    {
	      while(msgQueue.size()<=0)
		msgQueue.wait();
	    }
	  catch(Exception e)
	    {
	      e.printStackTrace();
	    }
	  aem=(AppEventMessage)msgQueue.elementAt(0);
	  msgQueue.removeElementAt(0);
	  return aem;
	}
    }
  protected void messageAvailable()
    {
    }
  public void run()
    {
      while(true)
	{
	  receiveAnyData();
	  synchronized(thsync)
	    {
	      if(running==false)
		{
		  th=null;
		  return;
		}
	    }
	}
    }
  
  public final boolean isMaster()
    // blocking; returns whether the user is the master of the session
    {
      return s_isMaster.request().isMaster();
    }

  public final String getUserName()
    // return the name of the current user;
    {
      return s_userName.request().getUsername();
    }

  public final String getMasterName()
    // returns the name of the current master
    {
      return s_masterName.request().getMastername();
    }

  public final String getHostname()
    // returns the dns name of the current machine
    {
      return s_hostName.request().getHostname();
    }

  public final Vector getParticipants()
    // returns the Vector of Strings naming all users in the session
    {
      return s_participants.request().getParticipants();
    }

  public final Vector getActiveUsers()
    // no idea what this returns
    {
      return s_activeUsers.request().getActiveUsers();
    }
  
  public Vector getLocalApps()
    // returns the list of all Tango applications running on the current
    // machine
    {
      return s_localApps.request().getLocalApps();
    }

  private boolean   v_isMaster;
  private Vector    v_participants;

  public final void receive(AppEventMessage aem)
    {
      if(aem.getType()==Const.DATA)
	{
	  synchronized(msgQueue)
	    {
	      msgQueue.addElement(aem);
	      msgQueue.notify();
	      messageAvailable();
	    }
	}
      else if(aem.getType()==Const.CONTROL)
	{
	  CAppAnswerEventMessage m=(CAppAnswerEventMessage)aem;
	  switch(m.getAnswerType())
	    {
	    case CAConsts.ISMASTER:
	      s_isMaster.response(m);
	      if(m.isMaster()!=v_isMaster)
		{
		  v_isMaster=!v_isMaster;
		  setMaster(v_isMaster);
		}
	      break;
	    case CAConsts.PARTICIPANTS:
	      s_participants.response(m) ;
	      if(v_participants==null)
		{
		  v_participants=m.getParticipants();
		  return;
		}
	      Object o1=new Integer(1);
	      Object o2=new Integer(2);
	      Object o3=new Integer(3);
	      Vector participants=m.getParticipants();
	      Hashtable h_parts=new Hashtable();
	      for(int i=0;i<participants.size();i++)
		{
		  h_parts.put(participants.elementAt(i),o1);
		}
	      for(int i=0;i<v_participants.size();i++)
		{
		  Object code=h_parts.get(v_participants.elementAt(i));
		  if(code==null)
		    h_parts.put(v_participants.elementAt(i),o2);
		  else
		    h_parts.put(v_participants.elementAt(i),o3);
		}
	      Enumeration e=h_parts.keys();
	      while(e.hasMoreElements())
		{
		  String key=(String)e.nextElement();
		  Object code=h_parts.get(key);
		  if(code==o1)
		    userEntered(key);
		  else if(code==o2)
		    userLeft(key);
		}
	      h_parts.clear();
	      v_participants=participants;
	      break;
	    case CAConsts.MASTERNAME:   s_masterName.response(m) ; break;
	    case CAConsts.USERNAME:     s_userName.response(m) ; break;
	    case CAConsts.HOSTNAME:     s_hostName.response(m) ; break;
	    case CAConsts.ACTIVEUSERS:  s_activeUsers.response(m) ; break;
	    case CAConsts.LOCALAPPS:    s_localApps.response(m) ; break;
	    }
	}
    }

  protected void userEntered(String name)
    // called if a user Enters the session
    {
    }
  protected void userLeft(String name)
    // called if a user leaves the session
    {
    }
  protected void setMaster(boolean isMaster)
    // called if the master changes
    {
    }
 
  public MessageOutputStream getMOS(Object o,boolean isRemovable)
    // returns a MessageOutputStream, if for TObject, the message autoatically will be
    // routed to the corresponding objects
    {
      try
	{
	  MessageOutputStream mos=new MessageOutputStream(this,isRemovable,new ByteArrayOutputStream());
	  mos.writeBoolean(isRemovable);
	  if(o instanceof TObject)
	    {
	      mos.writeByte(TOBJECT_DAT);
	      mos.writeUTF((String)trobject.get(o));
	    }
	  else
	    {
	      mos.writeByte(NON_TOBJECT_DAT);
	    }
	  return mos;
	}
      catch(Exception e)
	{
	  e.printStackTrace();
	}
      return null;
    }

  long lastSend;
  final void sendData(MessageOutputStream mos,boolean isRemovable)
    {
      if(isRemovable==false || System.currentTimeMillis()-lastSend>500)
	{
	  abi.send(new AppEventMessage(Const.DATA,mos.getByteStream().toByteArray()));
	  lastSend=System.currentTimeMillis();
	}
    }

   
  private final void receiveAnyData()
    {
      try
	{
	  AppEventMessage am=messageNext(); 
	  ByteArrayInputStream bis=new ByteArrayInputStream(am.getData());
	  DataInputStream dis=new DataInputStream(bis);
	  boolean isRemovable=dis.readBoolean();
	  if(isRemovable && msgQueue.size()>1)
	    return;
	  byte b=dis.readByte();
	  switch(b)
	    {
	    case TOBJECT_CRE:
	      {
		String is=dis.readUTF();
		String oname=dis.readUTF();
		Class c=Class.forName(oname);
		TObject oto=(TObject)tobject.get(is);
		if(oto==null)
		  {
		    Class cls[]={this.getClass(),Class.forName("java.io.DataInputStream")};
		    Constructor cn=c.getConstructor(cls);
		    Object tb[]={this,dis};
		    TObject to=(TObject)cn.newInstance(tb);
		    tobject.put(is,to);
		    trobject.put(to,is);
		  }
		else
		  oto.update(dis);
		break;
	      }
	    case TOBJECT_DEST:
	      {
		String is=dis.readUTF();
		TObject to=(TObject)tobject.get(is);
		tobject.remove(is);
		if(to!=null)
		  to.dispose();
		break;
	      }
	    case TOBJECT_DAT:
	      {
		String is=dis.readUTF();
		TObject to=(TObject)tobject.get(is);
		to.receive(dis);
		break;
	      }
	    case NON_TOBJECT_DAT:
	      {
		receiveData(dis);
		break;
	      }
	    case INIT_REQ:
	      {
		if(isMaster())
		  {
		    MessageOutputStream mos=new MessageOutputStream(this,false,new ByteArrayOutputStream());
		    mos.writeBoolean(false);
		    mos.writeByte(INIT_RES);
		
		    ByteArrayOutputStream bos;
		    DataOutputStream dos;

		    // sending general non-object specific initialization info
		    bos=new ByteArrayOutputStream();
		    dos=new DataOutputStream(bos);
		    writeInit(dos);
		    mos.writeInt(bos.size());
		    bos.writeTo(mos);

		    // sending info needed to create all objects
		    Enumeration e=tobject.keys();
		    while(e.hasMoreElements())
		      {
			String key=(String)e.nextElement();
			TObject to=(TObject)tobject.get(key);
			bos=new ByteArrayOutputStream();
			dos=new DataOutputStream(bos);
			dos.writeUTF(key);
			dos.writeUTF(to.getClass().getName());
			to.writeTo(dos);
			mos.writeInt(bos.size());
			bos.writeTo(mos);
		      }
		    mos.send();
		  }
		break;
	      }
	    case INIT_RES:
	      {
		if(initialized==false)
		  {
		    int len;
		    byte barray[];
		    DataInputStream ldis;

		    // reading general initialization info
		    len=dis.readInt();
		    barray=new byte[len];
		    dis.readFully(barray);
		    ldis=new DataInputStream(new ByteArrayInputStream(barray));
		    readInit(ldis);
	    
		    while(dis.available()>0)
		      {
			len=dis.readInt();
			barray=new byte[len];
			dis.readFully(barray);
			ldis=new DataInputStream(new ByteArrayInputStream(barray));
			String key=ldis.readUTF();
			String cname=ldis.readUTF();

			Class  c=Class.forName(cname);
			TObject to;
		    
			to=(TObject)tobject.get(key);
			if(to==null)
			  {
			    Class cls[]={this.getClass(),Class.forName("java.io.DataInputStream")};
			    Constructor cns[]=c.getConstructors();
			    Constructor cn=c.getConstructor(cls);
			    Object tb[]={this,ldis};
			    to=(TObject)cn.newInstance(tb);
			    tobject.put(key,to);
			    trobject.put(to,key);
			  }
			else
			  {
			    if(to.getClass().equals(c))
			      to.update(ldis);
			  }
		      }
		    initialized=true;
		    synchronized(this)
		      {
			notifyAll();
		      }
		  }
		break;
	      }
	    default:
	      throw new Exception("Unable to interpret the data");
	    }
	}
      catch(Exception e)
	{
	  e.printStackTrace();
	}
    }
  protected void receiveData(DataInputStream dis) throws Exception
    // called if received non TObject related data
    {
    }
  protected void writeInit(DataOutputStream dos) throws Exception
    // called when sending state to another application. This method is called before
    // sending all TObjects.
    {
    }
  protected void readInit(DataInputStream dis) throws Exception
    // called upon receiving of state of the session. This is called before creation of
    // TObjects.
    {
    }

  public final void add(TObject to)
    {
      try
	{
	  String id=newId();
	  tobject.put(id,to);
	  trobject.put(to,id);
	  MessageOutputStream mos=new MessageOutputStream(this,false,new ByteArrayOutputStream());
	  mos.writeBoolean(false);
	  mos.writeByte(TOBJECT_CRE);
	  mos.writeUTF((String)trobject.get(to));
	  mos.writeUTF(to.getClass().getName());
	  to.writeTo(mos);
	  mos.send();
	}
      catch(Exception e)
	{
	  e.printStackTrace();
	}
    }
  public final void add(String key,TObject to)
    {
      try
	{
	  TObject oto=(TObject)tobject.get(key);
	  if(oto!=null)
	    {
	      if(oto!=to)
		{
		  remove(oto);
		  oto.dispose();
		}
	      else
		return;
	    }
	  tobject.put(key,to);
	  trobject.put(to,key);
	  if(initialized)
	    {
	      MessageOutputStream mos=new MessageOutputStream(this,false,new ByteArrayOutputStream());
	      mos.writeBoolean(false);
	      mos.writeByte(TOBJECT_CRE);
	      mos.writeUTF((String)trobject.get(to));
	      mos.writeUTF(to.getClass().getName());
	      to.writeTo(mos);
	      mos.send();
	    }
	}
      catch(Exception e)
	{
	}
    }
  
  public final void remove(TObject to)
    {
      try
	{
	  Object key=trobject.get(to);
	  tobject.remove(key);
	  trobject.remove(to);
	  MessageOutputStream mos=new MessageOutputStream(this,false,new ByteArrayOutputStream());
	  mos.writeBoolean(false);
	  mos.writeByte(TOBJECT_DEST);
	  mos.writeUTF((String)key);
	  mos.send();
	}
      catch(Exception e)
	{
	  e.printStackTrace();
	}
    }
  public TObject getTObject(String id)
    {
      return (TObject)tobject.get(id);
    }
  public String getKey(TObject to)
    {
      return (String)trobject.get(to);
    }
  private static int k=0;
  public  String newId()
    {
      String s=getUserName()+"\n"+k;
      k++;
      return s;
    }

  public void dispose()
    {
      stop();
    }
  public final void stop()
    {
      synchronized(thsync)
	{	
	  running=false;
	}
    }
  public final void start()
    {
      synchronized(thsync)
	{
	  running=true;
	  if(th==null)
	    {
	      th=new Thread(this);
	      th.start();   
	    }
	}
    }
  public final void waitForInitialization()
    {
      synchronized(this)
	{
	  if(!initialized)
	    {
	      try
		{
		  wait();
		}
	      catch(Exception e)
		{
		  e.printStackTrace();
		}
	    }
	}
    }
}






